@@ -15,7 +15,11 @@ setClass( "CFuncList", contains = "list" )
 cfunction <- function(sig=character(), body=character(), includes=character(), otherdefs=character(),
                       language=c("C++", "C", "Fortran", "F95", "ObjectiveC", "ObjectiveC++"),
                       verbose=FALSE, convention=c(".Call", ".C", ".Fortran"), Rcpp=FALSE,
-                      cppargs=character(), cxxargs=character(), libargs=character()) {
+                      cppargs=character(), cxxargs=character(), libargs=character(),
+                      dim = NULL, implicit = NULL, module = NULL) {
+
+ if (missing (convention) & !missing(language))
+      convention <- switch (EXPR = language, "Fortran" = ".Fortran", "F95" = ".Fortran", ".C" = ".C", ObjectiveC = ".Call", "ObjectiveC++" = ".Call", "C++" = ".Call")
 
   convention <- match.arg(convention)
 
@@ -35,10 +39,19 @@ cfunction <- function(sig=character(), body=character(), includes=character(), o
   if( length(sig) != length(body) )
     stop("mismatch between the number of functions declared in 'sig' and the number of function bodies provided in 'body'")
 
+  if (is.null(dim)) 
+    dim <- as.list(rep("(*)", length(sig)))
+  else {            # this assumes fortran style
+      if (!is.list(dim)) 
+        dim <- list(dim)
+      if (length(dim) != length(sig)) 
+        stop("mismatch between the number of functions declared in 'sig' and the number of dimensions declared in 'dim'")
+  }
+
   if (Rcpp) {
-      if (!require(Rcpp)) stop("Rcpp cannot be loaded, install it or use the default Rcpp=FALSE")
+      if (!requireNamespace("Rcpp", quietly=TRUE))
+          stop("Rcpp cannot be loaded, install it or use the default Rcpp=FALSE", call.=FALSE)
       cxxargs <- c(Rcpp:::RcppCxxFlags(), cxxargs)	# prepend information from Rcpp
-      libargs <- c(Rcpp:::RcppLdFlags(), libargs)	# prepend information from Rcpp
   }
   if (length(cppargs) != 0) {
       args <- paste(cppargs, collapse=" ")
@@ -55,7 +68,7 @@ cfunction <- function(sig=character(), body=character(), includes=character(), o
       if (verbose) cat("Setting PKG_LIBS to", args, "\n")
       Sys.setenv(PKG_LIBS=args)
   }
-
+  types <- vector(mode="list", length=length(sig))
   ## GENERATE THE CODE
   for ( i in seq_along(sig) ) {
     ## C/C++ with .Call convention *********************************************
@@ -102,11 +115,11 @@ cfunction <- function(sig=character(), body=character(), includes=character(), o
       }
   	  ## determine function header
   	  if ( length(sig[[i]]) > 0 ) {
-  	    types <- pmatch(sig[[i]], c("logical", "integer", "double", "complex",
+  	    types[[i]] <- pmatch(sig[[i]], c("logical", "integer", "double", "complex",
   	                       "character", "raw", "numeric"), duplicates.ok = TRUE)
-  	    if ( any(is.na(types)) ) stop( paste("Unrecognized type", sig[[i]][is.na(types)]) )
+  	    if ( any(is.na(types[[i]])) ) stop( paste("Unrecognized type", sig[[i]][is.na(types[[i]])]) )
   	    decls <- c("int *", "int *", "double *", "Rcomplex *", "char **",
-  	               "unsigned char *", "double *")[types]
+  	               "unsigned char *", "double *")[ types[[i]] ]
   	    funCsig <- paste(decls, names(sig[[i]]), collapse=", ")
 	    }
 	    else funCsig <- ""
@@ -121,7 +134,9 @@ cfunction <- function(sig=character(), body=character(), includes=character(), o
   	  code <- paste( code, "\n}\n", sep="")
     }
     ## .Fortran convention *****************************************************
-    else {
+    else {                                              
+     # old-style fortran requires 6 columns not used
+     lead <- ifelse (language == "Fortran", "      ","") 
   	  if (i == 1) {
 	      ## no default includes, include further includes
 	      code <- paste(includes, collapse="\n")
@@ -130,26 +145,43 @@ cfunction <- function(sig=character(), body=character(), includes=character(), o
       }
   	  ## determine function header
   	  if ( length(sig[[i]]) > 0 ) {
-  	    types <- pmatch(sig[[i]], c("logical", "integer", "double", "complex",
+  	    types[[i]] <- pmatch(sig[[i]], c("logical", "integer", "double", "complex",
   	                       "character", "raw", "numeric"), duplicates.ok = TRUE)
-  	    if ( any(is.na(types)) ) stop( paste("Unrecognized type", sig[[i]][is.na(types)]) )
-  	    if (6 %in% types) stop( "raw type unsupported by .Fortran()" )
+  	    if ( any(is.na(types[[i]])) ) stop( paste("Unrecognized type", sig[[i]][is.na(types[[i]])]) )
+  	    if (6 %in% types[[i]]) stop( "raw type unsupported by .Fortran()" )
   	    decls <- c("INTEGER", "INTEGER", "DOUBLE PRECISION", "DOUBLE COMPLEX",
-  	               "CHARACTER*255", "Unsupported", "DOUBLE PRECISION")[types]
-  	    decls <- paste("      ", decls, " ", names(sig[[i]]), "(*)", sep="", collapse="\n")
+  	               "CHARACTER*255", "Unsupported", "DOUBLE PRECISION")[ types[[i]] ]
+  	    decls <- paste(lead, decls, " ", names(sig[[i]]), dim[[i]], sep="", collapse="\n")
   	    funCsig <- paste(names(sig[[i]]), collapse=", ")
   	  }
   	  else {
 	      decls <- ""
 	      funCsig <- ""
 	    }
-  	  funCsig <- paste("      SUBROUTINE", names(sig)[i], "(", funCsig, ")\n", sep=" ")
+       	    
+  	  funCsig <- paste(lead,"SUBROUTINE", names(sig)[i], "(", funCsig, ")\n", sep=" ")
+      ## old-style FORTRAN line length restricted to 72 characters
+      if (language == "Fortran") {
+        if ((cl <- nchar(funCsig)) >= 72) {
+          fstring <- substr(funCsig, 72, cl)
+          funCsig <- substr(funCsig, 1, 71)
+          while ((cf <- nchar(fstring)) > 66) {
+            funCsig <- paste(funCsig, "\n     &", substr(fstring, 1, 66), sep = "")
+            fstring <- substr(fstring, 67, cf)
+          } 
+          if (cf > 0) funCsig <- paste(funCsig, "\n     &", fstring, sep = "")
+            funCsig <- paste(funCsig, "\n")   
+        } 
+      } 
+      ## IMPLICIT statement and module use
+      if (is.character(module)) funCsig <- paste(funCsig, lead, "USE ", module, "\n", sep = "")
+      if (is.character(implicit)) funCsig <- paste(funCsig, lead, "IMPLICIT ", implicit, "\n", sep = "")          
   	  ## OPEN function
-  	  code <- paste( code, funCsig, decls, collapse="\n")
+  	  code <- paste( code, funCsig, decls, "\n", collapse="\n", sep="")
   	  ## add code, split lines
   	  code <- paste( code, paste(body[[i]], collapse="\n"), sep="")
   	  ## CLOSE function
-  	  code <- paste( code, "\n      RETURN\n      END\n", sep="")
+  	  code <- paste( code, "\n", lead, "RETURN\n", lead, "END\n\n", sep="")
     }
   } ## for along signatures
 
@@ -188,14 +220,20 @@ cfunction <- function(sig=character(), body=character(), includes=character(), o
     ## create .C/.Call function call that will be added to 'fn'
     if (convention == ".Call") {
       body <- quote( CONVENTION("EXTERNALNAME", ARG) )[ c(1:2, rep(3, length(sig[[i]]))) ]
-      for ( j in seq(along = sig[[i]]) ) body[[j+2]] <- as.name(names(sig[[i]])[j])
+      for ( j in seq_along(sig[[i]]) ) body[[j+2]] <- as.name(names(sig[[i]])[j])
     }
     else {
       body <- quote( CONVENTION("EXTERNALNAME", as.logical(ARG), as.integer(ARG),
                     as.double(ARG), as.complex(ARG), as.character(ARG),
-          			    as.raw(ARG), as.double(ARG)) )[ c(1:2,types+2) ]
+          			    as.raw(ARG), as.double(ARG)) )[ c(1:2,types[[i]]+2) ]
       names(body) <- c( NA, "", names(sig[[i]]) )
-      for ( j in seq(along = sig[[i]]) ) body[[j+2]][[2]] <- as.name(names(sig[[i]])[j])
+      for ( j in seq_along(sig[[i]]) ) body[[j+2]][[2]] <- as.name(names(sig[[i]])[j])
+## OLD VERSION -- does not work for lists of functions
+#      body <- quote( CONVENTION("EXTERNALNAME", as.logical(ARG), as.integer(ARG),
+#                    as.double(ARG), as.complex(ARG), as.character(ARG),
+#          			    as.raw(ARG), as.double(ARG)) )[ c(1:2,types+2) ]
+#      names(body) <- c( NA, "", names(sig[[i]]) )
+#      for ( j in seq(along = sig[[i]]) ) body[[j+2]][[2]] <- as.name(names(sig[[i]])[j])
     }
     body[[1]] <- get(convention)
     body[[2]] <- getNativeSymbolInfo( names(sig)[i], DLL )$address
@@ -252,21 +290,23 @@ compileCode <- function(f, code, language, verbose) {
 
   setwd(dirname(libCFile))
   errfile <- paste( basename(libCFile), ".err.txt", sep = "" )
-  cmd <- paste(R.home(component="bin"), "/R CMD SHLIB ", basename(libCFile), " 2> ", errfile, sep="")
-  if (verbose) cat("Compilation argument:\n", cmd, "\n")
-  compiled <- system(cmd, intern=!verbose)
+  cmd <- paste0(R.home(component="bin"), "/R")
+  if ( verbose ) system2(cmd, args = paste(" CMD SHLIB --dry-run", basename(libCFile)))
+  compiled <- system2(cmd, args = paste(" CMD SHLIB", basename(libCFile)),
+                      stdout = FALSE, stderr = errfile)
   errmsg <- readLines( errfile )
   unlink( errfile )
-  writeLines( errmsg )
-  setwd(wd)
 
   if ( !file.exists(libLFile) && file.exists(libLFile2) ) libLFile <- libLFile2
   if ( !file.exists(libLFile) ) {
     cat("\nERROR(s) during compilation: source code errors or compiler configuration errors!\n")
+    if ( !verbose ) system2(cmd, args = paste(" CMD SHLIB --dry-run --preclean", basename(libCFile)))
     cat("\nProgram source:\n")
     code <- strsplit(code, "\n")
     for (i in 1:length(code[[1]])) cat(format(i,width=3), ": ", code[[1]][i], "\n", sep="")
-    stop( paste( "Compilation ERROR, function(s)/method(s) not created!", paste( errmsg , collapse = "\n" ) ) )
+    cat("\nCompilation ERROR, function(s)/method(s) not created!\n")
+    if ( sum(nchar(errmsg)) > getOption("warning.length") ) stop(tail(errmsg))
+    else stop(errmsg)
   }
   return( libLFile )
 }
